// x.c
// (C) 1998 lars brinkhoff

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>

#include "pix.h"

struct info
{
  struct Options *options;
  Display *display;
  int screen;
  int depth;
  Visual *visual;
  Window window;
  GC gc;
  GC colors[256];
  Pixmap image;
};

static const int D[4][4] =
{ {  1,  9,  3, 12 },
  { 15,  5, 13,  7 },
  {  4, 11,  2, 10 },
  { 14,  8, 16,  6 } };

void pixel_x (void *info_, int x, int y, double r, double g, double b)
{
  struct info *info = info_;
  double i;
  int u, v;
  int col;
  
  i = 0.3 * r + 0.6 * b + 0.1 * b;
  
  for (u = 0; u < info->options->scale; u++)
    {
      for (v = 0; v < info->options->scale; v++)
	{
	  int X, Y;
	  
	  X = info->options->scale * x + u;
	  Y = info->options->scale * y + v;

	  if (info->depth == 1)
	    //col = i < ((double)rand () / RAND_MAX) ? 0 : 1;
	    //col = 17 * i > D[X % 4][Y % 4] ? 1 : 0;
	    col = i + (double)rand () / RAND_MAX / 17.0 - 0.5 / 17 >
		  D[X % 4][Y % 4] / 17.0 ? 1 : 0;
	  else
	    col = 255 * i + .5;
	  XDrawPoint (info->display,
		      info->image,
		      info->colors[col],
		      X, Y);
	}
    }
  
  if (info->options->interleave > 1 &&
      x % info->options->interleave == info->options->interleave - 1)
    {
      if (y % info->options->interleave == info->options->interleave - 1)
	{
	  int x1 = x - info->options->interleave + 1;
	  int y1 = y - info->options->interleave + 1;
	  XCopyArea (info->display,
		     info->image,
		     info->window,
		     info->gc,
		     info->options->scale * x1, info->options->scale * y1,
		     info->options->interleave * info->options->scale,
		     info->options->interleave * info->options->scale,
		     info->options->scale * x1, info->options->scale * y1);
	  XFlush (info->display);
	}
    }
  else if (x + 1 == info->options->width)
    {
      XCopyArea (info->display,
		 info->image,
		 info->window,
		 info->gc,
		 0, info->options->scale * y,
		 info->options->width * info->options->scale,
		 info->options->scale,
		 0, info->options->scale * y);
      XFlush (info->display);
    }
}

void wait_x (void *info_)
{
  struct info *info = info_;
  XEvent evnt;
  
  XSelectInput(info->display, info->window,
	       KeyPressMask | KeyReleaseMask | ExposureMask);

  for (;;)
    {
      XCopyArea (info->display,
		 info->image,
		 info->window,
		 info->gc,
		 0, 0,
		 info->options->width * info->options->scale,
		 info->options->height * info->options->scale,
		 0, 0);
      XFlush (info->display);
      XNextEvent (info->display, &evnt);
      switch(evnt.type)
	{
	case KeyPress:
	  exit(0);
	}
    }
}

void *
init_x (struct Options *options)
{
  XWindowAttributes attr;
  struct info *info;
  XGCValues values;
  char tmpcol[20];
  XColor a, b;
  int i;
  
  info = malloc (sizeof (struct info));
  if (info == 0)
    return 0;
  info->options = options;

  info->display = XOpenDisplay (0);
  if (info->display == 0)
    return 0;

  options->wait = wait_x;
  options->pixel = pixel_x;

  info->screen = DefaultScreen (info->display);
  info->depth = DefaultDepth (info->display,
			      info->screen);
  info->visual = DefaultVisual (info->display,
				info->screen);
  info->window = XCreateSimpleWindow (info->display,
				      RootWindow (info->display,
						  info->screen),
				      0, 0,
				      options->width * options->scale,
				      options->height * options->scale,
				      4, 0, 0);
  
  values.foreground = 0;
  values.plane_mask = 0x00ff;
  values.graphics_exposures = 0;
  info->gc = XCreateGC (info->display,
			info->window,
			GCForeground | GCPlaneMask |
			GCGraphicsExposures, 
			&values);
  
  XMapWindow (info->display, info->window);
  
  info->image = XCreatePixmap (info->display,
			       info->window,
			       options->width * options->scale,
			       options->height * options->scale,
			       info->depth);
  
  XGetWindowAttributes (info->display,
			info->window, &attr);
  
  for (i= 0; i < 1 << info->depth; i++)
    {
      int z;
      
      z = 256 >> info->depth;
      sprintf (tmpcol, "#%02X%02X%02X", i*z, i*z, i*z);
      XAllocNamedColor (info->display, attr.colormap, tmpcol, &a, &b);
      values.foreground = a.pixel;
      info->colors[i] = XCreateGC (info->display,
				   info->window,
				   GCForeground, &values);
    }

  return info;
}
